;----------------------------------------------------------------------------
; Bock's Birthday 2011
; (C) 2011 Charles MacDonald
; WWW: http://cgfm2.emuviews.com
;----------------------------------------------------------------------------

                        .include "global.inc"
                        .include "macros.inc"
                        .include "init.inc"
                        .include "sound.inc"
                        .include "vdp.inc"
                        .include "joy.inc"

;------------------------------------------------------------------------------
; Variables
;------------------------------------------------------------------------------

                        .ENUM   VAR_BASE
SLOT_SELECT             ds      1       ; What the BIOS puts here
VDP_STATUS              ds      1       ; VDP status during an interrupt
FRAME_COUNT             ds      1       ; Frame counter (60Hz)
PAUSE_FLAG              ds      1       ; Pause flag
JOY_NEW                 ds      1       ; Current 1P joypad state
JOY_OLD                 ds      1       ; Previous 1P joypad state
JOY_DELTA               ds      1       ; New keypresses for 1P joypad

sprite_xpos             ds      1       ; X coordinate of sprite
sprite_ypos             ds      1       ; Y coordinate of sprite
toggle                  ds      1       ; Address toggle per line interrupt
slomo_counter           ds      1       ; Slow motion counter
                        .ENDE

.define CAKE_ZOOM       0
.define CAKE_COLW       $08*(CAKE_ZOOM+1)
.define CAKE_COLH       $10*(CAKE_ZOOM+1)
.define CAKE_XADJ       $20             ; X position adjustment
.define CAKE_YADJ       $20             ; Y position adjustment
.define CAKE_XSCALE     (1+CAKE_ZOOM)
.define CAKE_YSCALE     1

.define                 SLOMO   0       ; 1= Enable slow motion for debug
.define                 SLOCNT  4       ; # of frames to wait per update
.define                 TOGGLE  0       ; 1= Swap SG's per line

; VRAM mapping
.define CAKE1_BASE      $0000   ; 0000-0FFF : 4K 
.define COLBUF_BASE     $1000   ; 1000-1FFF : 4K 
.define CAKE2_BASE      $2000   ; 2000-2FFF : 4K 
.define FREE_VRAM       $3000   ; 3000-37FF : 2K
.define NT_BASE         $3800   ; 3800-3EFF : 1792 bytes (32x28 words)
.define SAT_BASE        $3F00   ; 3F00-3FFF : 256 bytes

;------------------------------------------------------------------------------
; Main program
;------------------------------------------------------------------------------

main:
                        ; Setup
                        call    setup_registers
                        call    setup_background
                        call    setup_sprites

                        ; Turn screen on
                        ld      hl, $8014
                        rst     10h
                        ld      hl, $8A00
                        rst     10h
                        ld      hl, $8120|2|CAKE_ZOOM
                        rst     10h

                        ; Prepare for raster effects
                        call    raster_prep

                        xor     a
                        .ifeq   SLOMO 1
                        ld      (slomo_counter), a
                        .endif
                        ld      (FRAME_COUNT), a

                        ; Enable interrupts
                        ei

                        ; Main loop
        mainloop:
                        ; To reduce interrupt latency
                        .rept   256
                        halt
                        .endr
                        jp      mainloop

;----------------------------------------------------------------------------
; Frame interrupt routine
;----------------------------------------------------------------------------

frame_routine:
                        ; Bump frame count
                        ld      a, (FRAME_COUNT)
                        inc     a
                        ld      (FRAME_COUNT), a

                        ; Enable display after running for eight frames
                        ; This ensures all eight background columns
                        ; have been drawn.
                        cp      $08
                        jr      nz, skip_screen_enable
                        ld      hl, $8160|2|CAKE_ZOOM
                        rst     10h
skip_screen_enable:
                        ; Update animation
                        call    advance_animation

                        ; Get ready for raster effects
                        call    raster_prep

                        .ifeq   TOGGLE 1
                        ld      a, (toggle)
                        xor     4
                        ld      (toggle), a
                        .endif

                        ret

frame_interrupt:

                        .ifeq   SLOMO 1
                        ld      a, (slomo_counter)
                        inc     a
                        ld      (slomo_counter), a
                        cp      SLOCNT
                        jr      nz, skip_frame
                        .endif

                        call    frame_routine

                        .ifeq   SLOMO 1
                        xor     a
                        ld      (slomo_counter), a
        skip_frame:
                        .endif
                        
                        ex      af, af'
                        ei
                        reti

;----------------------------------------------------------------------------
; Line interrupt routine
;----------------------------------------------------------------------------

line_interrupt:
                        exx

                        .ifeq   TOGGLE 1
                        ld      a, (toggle)
                        xor     4
                        ld      (toggle), a
                        out     ($BF), a
                        ld      a, $86
                        out     ($BF), a
                        .endif

                        exx
                        ex      af, af'
                        ei
                        reti

; New frame - prepare for raster effects
raster_prep:
                        exx

                        exx
                        ret

;----------------------------------------------------------------------------
; NMI handler
;----------------------------------------------------------------------------

nmi_handler:
                        ; Set pause flag
                        ld      a, $01
                        ld      (PAUSE_FLAG), a
                        ret

;----------------------------------------------------------------------------
; Subroutines
;----------------------------------------------------------------------------

; Set up sprites
setup_sprites:
                        ; Set initial sprite position
                        ld      a, $00
                        ld      (sprite_xpos), a
                        ld      a, $00
                        ld      (sprite_ypos), a
                        ld      a, $FB
                        ld      (toggle), a

                        ; Load primary cake tile set
                        ld      hl, $4000|CAKE1_BASE
                        rst     10h
                        ld      hl, cake_sprite_pattern1
                        call    load_4k_block

                        ; Load alternate cake tile set
                        ld      hl, $4000|CAKE2_BASE
                        rst     10h
                        ld      hl, cake_sprite_pattern2
                        call    load_4k_block
                        ret

; Load 4K of data into VRAM
load_4k_block:
                        ld      b, $00
        load_vram_data: .rept    16
                        ld      a, (hl)
                        inc     hl
                        out     ($BE), a
                        .endr
                        djnz    load_vram_data
                        ret

advance_animation:
                        call    update_background
                        call    update_palette
                        call    update_sprites
                        ret

update_sprites:
                        ; Calculate X,Y position of sprites
                        ld      a, (FRAME_COUNT)
                        ld      l, a
                        ld      h, (sine256>>8)&$FF
                        ld      a, (hl)
                        .rept   CAKE_XSCALE
                        srl     a
                        .endr
                        sub     CAKE_YADJ
                        ld      (sprite_ypos), a

                        ld      a, l
                        add     a, a  
                        add     a, $80
                        ld      l, a
                        ld      a, (hl)
                        .rept   CAKE_YSCALE
                        srl     a
                        .endr
                        add     a, CAKE_XADJ
                        ld      (sprite_xpos), a

                        call    update_sprite_ypos
                        call    update_sprite_xpos
                        ret

update_sprite_ypos:
                        ; Point to sprite table Y area
                        ld      hl, $4000|SAT_BASE
                        rst     10h

                        ; Update Y positions for each row of 8 sprites
                        ; Skip setting a Y position of $D0 which terminates
                        ; sprite rendering
                        ;
                        ld      a, (sprite_ypos)
                        ld      bc, $08BE
        update_ypos:    cp      $D0
                        ret     z
                        .rept   8
                        out     ($BE), a
                        .endr
                        add     a, CAKE_COLH
                        djnz    update_ypos
                        ret

update_sprite_xpos:
                        ; Point to sprite table X,N area
                        ld      hl, $4080|SAT_BASE
                        rst     10h

                        ; Update X positions for each row of 8 sprites
                        ld      de, $0000
                        ld      bc, $08BE
        update_xnpos2:
                        ld      a, (sprite_xpos)
                        .rept   8
                        out     ($BE), a
                        add     a, CAKE_COLW
                        out     (c), e
                        inc     e       ; Use even tiles for 8x16 sprites
                        inc     e
                        .endr
                        djnz    update_xnpos2

                        ret

update_background:
                        ; Set vertical scroll
                        ld      a, (FRAME_COUNT)
                        out     ($BF), a
                        ld      a, $89
                        out     ($BF), a

                        ; Set horizontal scroll
                        ; XScroll = (frame_count ^ 7) << 5;
                        ld      a, (FRAME_COUNT)
                        rrca
                        rrca
                        rrca
                        and     $E0
                        xor     $E0
                        out     ($BF), a
                        ld      a, $88
                        out     ($BF), a

                        ; Set address of column tile buffer to update
                        ; VAddr = $4000 = (frame_count & 7) << 8;
                        xor     a
                        out     ($BF), a
                        ld      a, (FRAME_COUNT)
                        and     $07
                        add     a, a
                        ld      b, a
                        ld      l, a
                        or      $40
                        or      (COLBUF_BASE>>8)&$FF
                        out     ($BF), a

                        ; Select address of animation frame to draw
                        ld      a, (FRAME_COUNT)
                        add     a, a
                        add     a, a
                        ld      l, a
                        ld      h, (sine256>>8)&$FF
                        ld      a, (hl)
                        srl     a
                        srl     a
                        and     $1F
                        add     a, a
                        add     a, (animation>>8)&$FF
                        ld      h, a
                        ld      l, 0

                        ; Load animation frame to VRAM
                        ld      c, $BE
                        xor     a
                        .rept   8*16*4
                        outi
                        .endr
                        ret

update_palette:
                        ; Set CRAM write address
                        xor     a
                        out     ($BF), a
                        ld      a, $C0
                        out     ($BF), a

                        ; Calculate offset in crossfade sequence
                        ld      a, (FRAME_COUNT)
                        srl     a
                        srl     a
                        srl     a
                        and     $1F
                        ld      l, a
                        ld      h, (xfade_seq>>8)&$FF
                        ld      a, (hl)

                        ; Calculate offset in crossfade table list
                        add     a, a
                        ld      l, a
                        ld      h, (xfade_tab>>8)&$FF
                        ld      e, (hl)
                        inc     l
                        ld      d, (hl)
                        ex      de, hl

                        ; Copy color ramp to CRAM
                        ld      bc, $10BE
                        .rept   16
                        outi
                        .endr

                        ret

;----------------------------------------------------------------------------
; Set up background
;----------------------------------------------------------------------------
setup_background:
                        ; Set up palette
                        ld      hl, $C000
                        rst     10h
                        ld      hl, cramtab
                        ld      bc, $20BE
                        otir

                        ; Load name table
                        ld      hl, $4000|NT_BASE
                        rst     10h
                        ld      hl, name_table
                        ld      bc, $70BE
        name_loop:      push    bc
                        .rept   8
                        outi
                        out     (c), 0
                        .endr
                        pop     bc
                        djnz    name_loop
                        ret

setup_registers:
                        ld      hl, $8004
                        rst     10h
                        ld      hl, $8100
                        rst     10h
                        ld      hl, $8700
                        rst     10h
                        ld      hl, $8100
                        rst     10h
                        ld      hl, $86FB
                        rst     10h
                        ld      hl, $8800
                        rst     10h
                        ld      hl, $8900
                        rst     10h

                        ; Set table addresses
                        setnt   NT_BASE
                        setsat  SAT_BASE

                        ret


;----------------------------------------------------------------------------
; OUTI/INI/LDI block routines
;----------------------------------------------------------------------------

                        defop   BLKSIZE, outi
outiblk:                ret

                        defop   BLKSIZE, ini
iniblk:                 ret  

                        defop   BLKSIZE, ldi
ldiblk:                 ret

                        defop   HALTSIZE, nop
haltblk:                ret

                        defop   NOPSIZE, nop
nopblk:                 ret
                       
                        .rept   OUT0SIZE
                        out     (c), 0
                        .endr
out0blk:                ret

;----------------------------------------------------------------------------
; Data (Fixed)
;----------------------------------------------------------------------------

; 0000-0FFF : Z80 code
; 1000-4FFF : cake.anm
; 5000-5FFF : cake1.vrm
; 6000-6FFF : cake2.vrm
; 7000-73FF : name.tbl (896 bytes)
; 7400      : Crossfade sequence
; 7500      : Sine table
; 7600      : Crossfade table and palette ramps
; 7700-7FFF : (Free space)

                        .orga   $1000 
animation:              .incbin "sim/cake.anm"

                        .orga   $5000 
cake_sprite_pattern1:    
                        .incbin "cake/cake1.vrm"
cake_sprite_pattern2:    
                        .incbin "cake/cake2.vrm"

                        .orga   $7000
name_table:             .incbin "name/name.tbl"

                        .orga   $7400
xfade_seq:
                        .db     0,0,0,0,0,0,1,1,2,2,3,3,3,3,3,3
                        .db     3,3,3,3,3,3,3,2,2,1,1,0,0,0,0,0

                        .orga   $7500
sine256:                .include "sine256.inc"

                        .orga   $7600
xfade_tab:
                        .dw     ramp+$00 ; Image 1
                        .dw     ramp+$10 ; Intermediate
                        .dw     ramp+$20 ; Intermediate
                        .dw     ramp+$30 ; Image 2

ramp:                   .incbin "ramp/ramp.bin"

cramtab:
                        ; Default background palette
                        .db     $00, $00, $3f, $3f
                        .db     $30, $30, $3f, $3f
                        .db     $03, $03, $3f, $3f
                        .db     $33, $33, $3f, $3f

                        ; PSP palette data macros

.macro                  defjasc
                        .db     ((\1>>4)&$3)|(\1&$0c)|((\1<<4)&$30)
                        .endm

.macro                  defhjasc
                        .db     ((\1>>5)&$3)|((\1>>1)&$0c)|((\1<<3)&$30)
                        .endm

                        ; Sprite palette (from PSP)
                        defjasc 0 
                        defjasc 0
                        defjasc 46
                        defjasc 57
                        defjasc 58
                        defjasc 60
                        defjasc 62
                        defjasc 63
                        defhjasc 0
                        defhjasc 0
                        defhjasc 46
                        defhjasc 57
                        defhjasc 58
                        defhjasc 60
                        defhjasc 62
                        defhjasc 63

;----------------------------------------------------------------------------
; SDSC tag
;----------------------------------------------------------------------------

.SDSCTAG 1.0, "Bock's Birthday 2011", "Happy Birthday Omar!", "Charles MacDonald"

;----------------------------------------------------------------------------
; End
;----------------------------------------------------------------------------
